@@ -34,7 +34,7 @@ extern int WIN32_SSL_password_prompt(tcn_pass_cb_t *data);
 #define ASN1_SEQUENCE 0x30
 #define ASN1_OID      0x06
 #define ASN1_STRING   0x86
-static int ssl_verify_OCSP(int ok, X509_STORE_CTX *ctx);
+static int ssl_verify_OCSP(X509_STORE_CTX *ctx);
 static int ssl_ocsp_request(X509 *cert, X509 *issuer, X509_STORE_CTX *ctx);
 #endif
 
@@ -52,8 +52,9 @@ static int ssl_ocsp_request(X509 *cert, X509 *issuer, X509_STORE_CTX *ctx);
  */
 static int SSL_app_data2_idx = -1;
 static int SSL_app_data3_idx = -1;
+static int SSL_app_data4_idx = -1;
 
-void SSL_init_app_data2_3_idx(void)
+void SSL_init_app_data_idx(void)
 {
     int i;
 
@@ -78,6 +79,15 @@ void SSL_init_app_data2_3_idx(void)
                                  "Third Application Data for SSL",
                                   NULL, NULL, NULL);
 
+    if (SSL_app_data4_idx > -1) {
+        return;
+    }
+
+    SSL_app_data4_idx =
+            SSL_get_ex_new_index(0,
+                                 "Fourth Application Data for SSL",
+                                  NULL, NULL, NULL);
+
 }
 
 void *SSL_get_app_data2(SSL *ssl)
@@ -102,6 +112,16 @@ void SSL_set_app_data3(SSL *ssl, void *arg)
     SSL_set_ex_data(ssl, SSL_app_data3_idx, arg);
 }
 
+void *SSL_get_app_data4(const SSL *ssl)
+{
+    return SSL_get_ex_data(ssl, SSL_app_data4_idx);
+}
+
+void SSL_set_app_data4(SSL *ssl, void *arg)
+{
+    SSL_set_ex_data(ssl, SSL_app_data4_idx, arg);
+}
+
 /* Simple echo password prompting */
 int SSL_password_prompt(tcn_pass_cb_t *data)
 {
@@ -310,8 +330,9 @@ int SSL_callback_SSL_verify(int ok, X509_STORE_CTX *ctx)
 #endif
 
     if (verify == SSL_CVERIFY_UNSET ||
-        verify == SSL_CVERIFY_NONE)
+        verify == SSL_CVERIFY_NONE) {
         return 1;
+    }
 
     if (SSL_VERIFY_ERROR_IS_OPTIONAL(errnum) &&
         (verify == SSL_CVERIFY_OPTIONAL_NO_CA)) {
@@ -348,7 +369,7 @@ int SSL_callback_SSL_verify(int ok, X509_STORE_CTX *ctx)
             ok = 0;
         }
         else {
-            int ocsp_response = ssl_verify_OCSP(ok, ctx);
+            int ocsp_response = ssl_verify_OCSP(ctx);
             if (ocsp_response == OCSP_STATUS_REVOKED) {
                 ok = 0 ;
                 errnum = X509_STORE_CTX_get_error(ctx);
@@ -507,7 +528,7 @@ int SSL_callback_alpn_select_proto(SSL* ssl, const unsigned char **out, unsigned
 #ifdef HAVE_OCSP_STAPLING
 
 /* Function that is used to do the OCSP verification */
-static int ssl_verify_OCSP(int ok, X509_STORE_CTX *ctx)
+static int ssl_verify_OCSP(X509_STORE_CTX *ctx)
 {
     X509 *cert, *issuer;
     int r = OCSP_STATUS_UNKNOWN;
@@ -722,24 +743,22 @@ static char **decode_OCSP_url(ASN1_OCTET_STRING *os, apr_pool_t *p)
 
 
 /* stolen from openssl ocsp command */
-static int add_ocsp_cert(OCSP_REQUEST **req, X509 *cert, X509 *issuer,
-                         STACK_OF(OCSP_CERTID) *ids)
+static int add_ocsp_cert(OCSP_REQUEST *req, X509 *cert, X509 *issuer)
 {
     OCSP_CERTID *id;
 
     if (!issuer)
         return 0;
-    if (!*req)
-        *req = OCSP_REQUEST_new();
-    if (!*req)
-        return 0;
     id = OCSP_cert_to_id(NULL, cert, issuer);
-    if (!id || !sk_OCSP_CERTID_push(ids, id))
+    if (!id)
         return 0;
-    if (!OCSP_request_add0_id(*req, id))
+    if (!OCSP_request_add0_id(req, id)) {
+        OCSP_CERTID_free(id);
         return 0;
-    else
+    } else {
+        /* id will be freed by OCSP_REQUEST_free() */
         return 1;
+    }
 }
 
 
@@ -801,7 +820,6 @@ static int ocsp_send_req(apr_socket_t *sock, BIO *req)
     int len;
     char buf[TCN_BUFFER_SZ];
     apr_status_t rv;
-    int ok = 1;
 
     while ((len = BIO_read(req, buf, sizeof buf)) > 0) {
         char *wbuf = buf;
@@ -815,12 +833,11 @@ static int ocsp_send_req(apr_socket_t *sock, BIO *req)
         } while (rv == APR_SUCCESS && remain > 0);
 
         if (rv != APR_SUCCESS) {
-            apr_socket_close(sock);
-            ok = 0;
+            return 0;
         }
     }
 
-    return ok;
+    return 1;
 }
 
 
@@ -909,7 +926,7 @@ err:
 /* Reads the response from the APR socket to a buffer, and parses the buffer to
    return the OCSP response  */
 #define ADDLEN 512
-static OCSP_RESPONSE *ocsp_get_resp(apr_socket_t *sock)
+static OCSP_RESPONSE *ocsp_get_resp(apr_pool_t *mp, apr_socket_t *sock)
 {
     int buflen;
     apr_size_t totalread = 0;
@@ -919,7 +936,7 @@ static OCSP_RESPONSE *ocsp_get_resp(apr_socket_t *sock)
     apr_pool_t *p;
     OCSP_RESPONSE *resp;
 
-    apr_pool_create(&p, NULL);
+    apr_pool_create(&p, mp);
     buflen = ADDLEN;
     buf = apr_palloc(p, buflen);
     if (buf == NULL) {
@@ -959,34 +976,29 @@ static OCSP_RESPONSE *ocsp_get_resp(apr_socket_t *sock)
 }
 
 /* Creates and OCSP request and returns the OCSP_RESPONSE */
-static OCSP_RESPONSE *get_ocsp_response(X509 *cert, X509 *issuer, char *url)
+static OCSP_RESPONSE *get_ocsp_response(apr_pool_t *p, X509 *cert, X509 *issuer, char *url)
 {
     OCSP_RESPONSE *ocsp_resp = NULL;
     OCSP_REQUEST *ocsp_req = NULL;
     BIO *bio_req;
     char *hostname, *path, *c_port;
     int port, use_ssl;
-    STACK_OF(OCSP_CERTID) *ids = NULL;
     int ok = 0;
     apr_socket_t *apr_sock = NULL;
     apr_pool_t *mp;
 
-    apr_pool_create(&mp, NULL);
-    ids = sk_OCSP_CERTID_new_null();
+    if (OCSP_parse_url(url,&hostname, &c_port, &path, &use_ssl) == 0 )
+        goto end;
 
-    /* problem parsing the URL */
-    if (OCSP_parse_url(url,&hostname, &c_port, &path, &use_ssl) == 0 ) {
-        sk_OCSP_CERTID_free(ids);
-        return NULL;
-    }
-
-    /* Create the OCSP request */
     if (sscanf(c_port, "%d", &port) != 1)
         goto end;
+
+    /* Create the OCSP request */
     ocsp_req = OCSP_REQUEST_new();
     if (ocsp_req == NULL)
-        return NULL;
-    if (add_ocsp_cert(&ocsp_req,cert,issuer,ids) == 0 )
+        goto end;
+
+    if (add_ocsp_cert(ocsp_req,cert,issuer) == 0 )
         goto free_req;
 
     /* create the BIO with the request to send */
@@ -995,29 +1007,30 @@ static OCSP_RESPONSE *get_ocsp_response(X509 *cert, X509 *issuer, char *url)
         goto free_req;
     }
 
+    apr_pool_create(&mp, p);
     apr_sock = make_socket(hostname, port, mp);
     if (apr_sock == NULL) {
-        ocsp_resp = NULL;
         goto free_bio;
     }
 
     ok = ocsp_send_req(apr_sock, bio_req);
-    if (ok)
-        ocsp_resp = ocsp_get_resp(apr_sock);
+    if (ok) {
+        ocsp_resp = ocsp_get_resp(mp, apr_sock);
+    }
+    apr_socket_close(apr_sock);
 
 free_bio:
     BIO_free(bio_req);
-
-free_req:
-    if(apr_sock && ok) /* if ok == 0 we have already closed the socket */
-        apr_socket_close(apr_sock);
-
     apr_pool_destroy(mp);
 
-    sk_OCSP_CERTID_free(ids);
+free_req:
     OCSP_REQUEST_free(ocsp_req);
 
 end:
+    OPENSSL_free(hostname);
+    OPENSSL_free(c_port);
+    OPENSSL_free(path);
+
     return ocsp_resp;
 }
 
@@ -1034,14 +1047,12 @@ static int process_ocsp_response(OCSP_RESPONSE *ocsp_resp, X509 *cert, X509 *iss
     r = OCSP_response_status(ocsp_resp);
 
     if (r != OCSP_RESPONSE_STATUS_SUCCESSFUL) {
-        OCSP_RESPONSE_free(ocsp_resp);
         return OCSP_STATUS_UNKNOWN;
     }
     bs = OCSP_response_get1_basic(ocsp_resp);
 
     certid = OCSP_cert_to_id(NULL, cert, issuer);
     if (certid == NULL) {
-        OCSP_RESPONSE_free(ocsp_resp);
         return OCSP_STATUS_UNKNOWN;
     }
     ss = OCSP_resp_get0(bs, OCSP_resp_find(bs, certid, -1)); /* find by serial number and get the matching response */
@@ -1057,7 +1068,7 @@ static int process_ocsp_response(OCSP_RESPONSE *ocsp_resp, X509 *cert, X509 *iss
 
     /* we clean up */
     OCSP_CERTID_free(certid);
-    OCSP_RESPONSE_free(ocsp_resp);
+    OCSP_BASICRESP_free(bs);
     return o;
 }
 
@@ -1087,7 +1098,7 @@ static int ssl_ocsp_request(X509 *cert, X509 *issuer, X509_STORE_CTX *ctx)
         int rv = OCSP_STATUS_UNKNOWN;
         /* for the time being just check for the fist response .. a better
            approach is to iterate for all the possible ocsp urls */
-        resp = get_ocsp_response(cert, issuer, ocsp_urls[0]);
+        resp = get_ocsp_response(p, cert, issuer, ocsp_urls[0]);
         if (resp != NULL) {
             rv = process_ocsp_response(resp, cert, issuer);
         } else {
@@ -1096,6 +1107,7 @@ static int ssl_ocsp_request(X509 *cert, X509 *issuer, X509_STORE_CTX *ctx)
         }
 
         if (resp != NULL) {
+            OCSP_RESPONSE_free(resp);
             apr_pool_destroy(p);
             return rv;
         }
